import json
import os
import random
from pathlib import Path

import pytest
import sqlalchemy.dialects
import sqlalchemy.exc
from sqlalchemy.orm import Session

import rasax.community.database.utils as db_utils
import rasax.community.sql_migrations as migrations
from rasax.community import utils, config
from rasax.community.database import Role
from tests.unit import conftest
from tests.unit.conftest import insert_test_user, TEST_USER


# Overwrite fixture to run per function
@pytest.fixture()
def session():
    session_maker = conftest.session_maker_for_tests()
    session = session_maker()

    migrations.run_migrations(session)

    # Insert general test user which can be used in later tests
    insert_test_user(session, TEST_USER)

    return session_maker()


@pytest.fixture
def all_versions():
    return [
        "0.21.0a3.dev45+geda689e",
        "0.21.0a3.dev8+gb7f2938",
        "0.21.0a2",
        "0.21.0a2.dev25+gc41e902",
        "0.21.0a2.dev2+gda67ff5",
        "0.21.0a1",
        "0.20.3.dev157+g02d6625",
        "0.20.3.dev1+g2d271d2",
        "0.20.2",
        "0.20.2a3.dev158+gd656601",
        "0.20.2a3.dev47+g8ab683f",
        "0.20.2a2",
        "0.20.2a2.dev60+gf23b110",
        "0.20.2a2.dev3+gf451a26",
        "0.20.2a1",
        "0.20.2.dev569+g1665f89",
        "0.20.2.dev26+gae2d791",
        "0.20.1",
        "0.20.1.dev157+g92c8166",
        "0.20.1.dev21+g26647d5",
        "0.20.0",
        "0.20.0a1",
        "0.19.6.dev336+gc771b11",
    ]


@pytest.mark.parametrize(
    "current_version, newer_stable_version",
    [
        ("0.21.0a4.dev22+nwow", "0.21.0a4.dev22+nwow"),
        ("0.19.6.dev336+gc771b11", "0.20.2"),
        ("0.20.3.dev1+g2d271d2", "0.20.3.dev1+g2d271d2"),
        ("0.20.2", "0.20.2"),
        ("0.21.0a3.dev45+geda689e", "0.21.0a3.dev45+geda689e"),
        ("0.20.1", "0.20.2"),
    ],
)
def test_get_newer_stable_version(all_versions, current_version, newer_stable_version):
    assert (
        utils._get_newer_stable_version(all_versions, current_version)
        == newer_stable_version
    )


def test_query_result_to_dict():
    query_result = ["zero", "one", "two", "three"]
    fields_query = [("d", True), ("a.c", True), ("a.b", True), ("b.c.d", True)]

    actual = utils.query_result_to_dict(query_result, fields_query)

    assert actual["d"] == query_result[0]
    assert actual["a"]["c"] == query_result[1]
    assert actual["a"]["b"] == query_result[2]
    assert actual["b"]["c"]["d"] == query_result[3]


def test_foreign_keys_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=True)
    with pytest.raises(sqlalchemy.exc.IntegrityError):
        db_connection.execute(f"DROP table {Role.__tablename__};")


def test_turn_off_foreign_keys_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=False)
    if isinstance(db_connection.dialect, sqlalchemy.dialects.sqlite.dialect):
        db_connection.execute(f"DROP table {Role.__tablename__};")


def test_turn_off_then_on_foreign_key_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=False)
    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=True)

    with pytest.raises(sqlalchemy.exc.IntegrityError):
        db_connection.execute(f"DROP table {Role.__tablename__};")


@pytest.mark.parametrize(
    "before_truncation, expected",
    [(1.0, 1.0), (1.1234, 1.12), (19231231.00, 19231231.00)],
)
def test_truncate_float(before_truncation: float, expected: float):
    assert utils.truncate_float(before_truncation, decimal_places=2) == expected


def test_add_plural_suffix():
    base_string = "found item"
    base_string_with_replacement_field = base_string + "{}"

    # list with single item yields singular case
    assert (
        utils.add_plural_suffix(base_string_with_replacement_field, [1]) == "found item"
    )

    # list with multiple items yields plural case
    assert (
        utils.add_plural_suffix(base_string_with_replacement_field, [1, 2, 3])
        == "found items"
    )


def test_query_result_len():
    count = random.randint(1, 10)
    result = utils.QueryResult([{} for _ in range(count)], count)
    assert len(result) == count


def test_decode_base_64():
    import base64

    expected = "hello"
    encoded = base64.b64encode(expected.encode("utf-8")).decode("utf-8")

    assert utils.decode_base64(encoded) == expected


@pytest.mark.parametrize(
    "original, encoded", [("testing", "dGVzdGluZw=="), ("hello", "aGVsbG8=")]
)
def test_encode_base_64(original, encoded):
    assert utils.encode_base64(original) == encoded


# noinspection PyProtectedMember
def test_db_query_from_environment():
    db_query_env_var = "DB_QUERY"

    # env var not set
    os.environ.pop(db_query_env_var, None)
    assert not db_utils._sql_query_parameters_from_environment()

    # env var correctly set
    query = {"driver": "some driver"}
    os.environ[db_query_env_var] = json.dumps(query)
    assert db_utils._sql_query_parameters_from_environment() == query

    # malformed env var
    malformed_db_query = '{"query": "badly formatted}'
    os.environ[db_query_env_var] = malformed_db_query
    assert not db_utils._sql_query_parameters_from_environment()


def test_get_project_directory_if_none_is_given() -> None:
    assert utils.get_project_directory() == Path()


def test_set_project_directory() -> None:
    directory = Path("test-directory")
    utils.set_project_directory(directory)

    assert utils.get_project_directory() == directory


def test_should_dump() -> None:
    with config.PROJECT_DIRECTORY.get_lock():
        config.PROJECT_DIRECTORY.value = "".encode("utf-8")

    assert not utils.should_dump()

    utils.set_project_directory("/tmp")

    assert utils.should_dump()


def test_write_text_to_file(tmpdir: Path) -> None:
    file_path = tmpdir / "sub" / "file.txt"

    assert not file_path.exists()

    utils.write_text_to_file(file_path, "some content")

    assert file_path.exists()
